/***********************************************************************
 *
 *  Gtk+ wrappers for GNU Smalltalk
 *
 ***********************************************************************/

/***********************************************************************
 *
 * Copyright 2001, 2003 Free Software Foundation, Inc.
 * Written by Paolo Bonzini, Norman Jordan, Mike S. Anderson.
 *
 * This file is part of GNU Smalltalk.
 *
 * GNU Smalltalk is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2, or (at your option) any later 
 * version.
 * 
 * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 ***********************************************************************/

#include "config.h"
#include "gstpub.h"

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <glib-object.h>
#include <glib.h>

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif

#include "blox-gtk.h"
#include "placer.h"

typedef struct SmalltalkClosure
{
  GClosure closure;
  OOP      receiver;
  OOP      selector;
  OOP      data;
  OOP      widget;
} SmalltalkClosure;

VMProxy *_blox_vm_proxy;
static GHashTable *h_callback = NULL;
static int pending_quit_count = 0;

static void blox_gtk_init ();
static void my_gtk_main_iteration ();
static void my_gtk_main_iteration_do (gboolean blocking);
static gboolean should_quit ();

static OOP convert_g_value_to_oop (const GValue *val);

static void fill_g_value_from_oop (GValue *val,
				   OOP oop);

static OOP container_get_child_property (GtkContainer *aParent,
					 GtkWidget *anObject,
					 const char *aProperty);

static void container_set_child_property (GtkContainer *aParent,
					  GtkWidget *anObject,
					  const char *aProperty,
					  OOP aValue);

static OOP widget_get_property (GObject *anObject,
				const char *aProperty);

static void widget_set_property (GObject *anObject,
				 const char *aProperty,
				 OOP aValue);

static void finalize_smalltalk_closure(gpointer      data,
				       GClosure     *closure);

static void invoke_smalltalk_closure(GClosure     *closure,
				     GValue       *return_value,
				     guint         n_param_values,
				     const GValue *param_values,
				     gpointer      invocation_hint,
				     gpointer      marshal_data);

static void connect_signal (OOP widget,
			    char *event_name,
			    OOP event_handler,
			    OOP user_data);
OOP
convert_g_value_to_oop (const GValue *val)
{
  GType type = G_VALUE_TYPE (val);
  GType fundamental;

  char v_char;
  gboolean v_boolean;
  gpointer v_ptr;
  long v_int;
  double v_float;

  if (G_TYPE_IS_FUNDAMENTAL (type))
    fundamental = type;
  else
    fundamental = G_TYPE_FUNDAMENTAL (type);

  switch (fundamental)
    {
    case G_TYPE_CHAR:
      v_char = g_value_get_char (val);
      return _blox_vm_proxy->charToOOP(v_char);

    case G_TYPE_BOOLEAN:
      v_boolean = g_value_get_boolean (val);
      return _blox_vm_proxy->boolToOOP(v_boolean);

    case G_TYPE_UCHAR:
      v_int = g_value_get_uchar (val);
      return _blox_vm_proxy->intToOOP(v_int);

    case G_TYPE_INT:
      v_int = g_value_get_int (val);
      return _blox_vm_proxy->intToOOP(v_int);

    case G_TYPE_UINT:
      v_int = g_value_get_uint (val);
      return _blox_vm_proxy->intToOOP(v_int);

    case G_TYPE_LONG:
      v_int = g_value_get_long (val);
      return _blox_vm_proxy->intToOOP(v_int);

    case G_TYPE_ULONG:
      v_int = g_value_get_ulong (val);
      return _blox_vm_proxy->intToOOP(v_int);

    case G_TYPE_ENUM:
      v_int = g_value_get_enum (val);
      return _blox_vm_proxy->intToOOP(v_int);

    case G_TYPE_FLAGS:
      v_int = g_value_get_flags (val);
      return _blox_vm_proxy->intToOOP(v_int);

    case G_TYPE_FLOAT:
      v_float = g_value_get_float (val);
      return _blox_vm_proxy->floatToOOP(v_float);

    case G_TYPE_DOUBLE:
      v_float = g_value_get_double (val);
      return _blox_vm_proxy->floatToOOP(v_float);

    case G_TYPE_STRING:
      v_ptr = (gpointer) g_value_get_string (val);
      return _blox_vm_proxy->stringToOOP(v_ptr);

    case G_TYPE_POINTER:
      v_ptr = g_value_get_pointer (val);
      return _blox_vm_proxy->cObjectToOOP(v_ptr);

    case G_TYPE_OBJECT:
    case G_TYPE_INTERFACE:
      /* Instantiate an appropriate subclass in the future... */
      v_ptr = g_value_get_object (val);
      if (fundamental == type
	  || G_TYPE_CHECK_INSTANCE_TYPE (v_ptr, type))
        {
          return _blox_vm_proxy->cObjectToOOP(v_ptr);
	}

    default:
      return NULL;
    }
}

void
fill_g_value_from_oop (GValue *return_value, OOP oop)
{
  GType type = G_VALUE_TYPE (return_value);
  GType fundamental;

  char v_char;
  gboolean v_boolean;
  gpointer v_ptr;
  long v_int;
  double v_float;

  if (G_TYPE_IS_FUNDAMENTAL (type))
    fundamental = type;
  else
    fundamental = G_TYPE_FUNDAMENTAL (type);

  switch (type)
    {
    case G_TYPE_NONE:
    case G_TYPE_INVALID:
      break;

    case G_TYPE_CHAR:
      v_char = _blox_vm_proxy->OOPToChar(oop);
      g_value_set_char (return_value, v_char);
      break;

    case G_TYPE_BOOLEAN:
      v_boolean = _blox_vm_proxy->OOPToBool(oop);
      g_value_set_boolean (return_value, v_boolean);
      break;

    case G_TYPE_UCHAR:
      v_int = _blox_vm_proxy->OOPToInt(oop);
      g_value_set_uchar (return_value, v_int);
      break;

    case G_TYPE_INT:
      v_int = _blox_vm_proxy->OOPToInt(oop);
      g_value_set_int (return_value, v_int);
      break;

    case G_TYPE_UINT:
      v_int = _blox_vm_proxy->OOPToInt(oop);
      g_value_set_uint (return_value, v_int);
      break;

    case G_TYPE_LONG:
      v_int = _blox_vm_proxy->OOPToInt(oop);
      g_value_set_long (return_value, v_int);
      break;

    case G_TYPE_ULONG:
      v_int = _blox_vm_proxy->OOPToInt(oop);
      g_value_set_ulong (return_value, v_int);
      break;

    case G_TYPE_ENUM:
      v_int = _blox_vm_proxy->OOPToInt(oop);
      g_value_set_enum (return_value, v_int);
      break;

    case G_TYPE_FLAGS:
      v_int = _blox_vm_proxy->OOPToInt(oop);
      g_value_set_flags (return_value, v_int);
      break;

    case G_TYPE_FLOAT:
      v_float = _blox_vm_proxy->OOPToFloat(oop);
      g_value_set_float (return_value, v_float);
      break;

    case G_TYPE_DOUBLE:
      v_float = _blox_vm_proxy->OOPToFloat(oop);
      g_value_set_double (return_value, v_float);
      break;

    case G_TYPE_STRING:
      v_ptr = _blox_vm_proxy->OOPToString(oop);
      g_value_set_string_take_ownership (return_value, v_ptr);
      break;

    case G_TYPE_POINTER:
    case G_TYPE_OBJECT:
    case G_TYPE_INTERFACE:
      v_ptr = _blox_vm_proxy->OOPToCObject(oop);
      if (fundamental == type
	  || G_TYPE_CHECK_INSTANCE_TYPE (v_ptr, type))
	{
          g_value_set_pointer (return_value, v_ptr);
          break;
	}

    default:
      fprintf (stderr, "Invalid type.\n");
      abort ();
    }
}

GClosure *
create_smalltalk_closure(OOP	     receiver,	
			 const char *selector,
			 OOP	     data,
			 OOP	     widget)
{
  GClosure *closure = g_closure_new_simple (sizeof (SmalltalkClosure), NULL);
  SmalltalkClosure *stc = (SmalltalkClosure *) closure;

  _blox_vm_proxy->registerOOP (receiver);
  _blox_vm_proxy->registerOOP (data);
  _blox_vm_proxy->registerOOP (widget);

  stc->receiver = receiver;
  stc->selector = _blox_vm_proxy->symbolToOOP(selector);
  stc->data = data;
  stc->widget = widget;

  g_closure_set_marshal (closure, invoke_smalltalk_closure);
  g_closure_add_finalize_notifier (closure, NULL, finalize_smalltalk_closure);
  return closure;
}

void
finalize_smalltalk_closure(gpointer      data,
			   GClosure     *closure)
{
  SmalltalkClosure *stc = (SmalltalkClosure *) closure;

  _blox_vm_proxy->unregisterOOP (stc->receiver);
  _blox_vm_proxy->unregisterOOP (stc->data);
  _blox_vm_proxy->unregisterOOP (stc->widget);
}

void
invoke_smalltalk_closure(GClosure     *closure,
			 GValue       *return_value,
			 guint         n_param_values,
			 const GValue *param_values,
			 gpointer      invocation_hint,
			 gpointer      marshal_data)
{
  OOP *args = alloca (sizeof (OOP) * (2 + n_param_values));
  SmalltalkClosure *stc = (SmalltalkClosure *) closure;
  OOP resultOOP;
  int i;

  args[0] = stc->widget;
  args[1] = stc->data;
  for (i = 0; i < n_param_values; i++)
    {
      OOP oop = convert_g_value_to_oop (&param_values[i]);
      if (!oop)
        {
          fprintf (stderr, "Invalid type, signal discarded.\n");

          if (return_value->g_type == G_TYPE_NONE)
	    return;
          else
	    abort ();
	}

      args[2 + i] = oop;
    }

  resultOOP = _blox_vm_proxy->nvmsgSend(stc->receiver, stc->selector, args,
					2 + n_param_values);

  if (return_value)
    fill_g_value_from_oop (return_value, resultOOP);
}

void
oop_for_widget (GObject *widget)
{
}

void
connect_signal (OOP widget, char *event_name, OOP event_handler,
		OOP user_data)
{
  GtkWidget *cWidget = _blox_vm_proxy->OOPToCObject (widget);
  const char *widgetClassName;
  const char *selector;
  GtkType currentType;
  GClosure *closure;

  currentType = GTK_OBJECT_TYPE (cWidget);
  widgetClassName = gtk_type_name (currentType);

  for (;;)
    {
      selector =
        g_hash_table_lookup (g_hash_table_lookup (h_callback, widgetClassName),
			     event_name);

      if (selector)
	break;

      if (strcmp (widgetClassName, "GtkObject") == 0)
	return;

      currentType = gtk_type_parent (currentType);
      widgetClassName = gtk_type_name (currentType);
    }

  closure = create_smalltalk_closure (event_handler, selector, user_data, widget);
  g_signal_connect_closure (cWidget, event_name, closure, FALSE);
}


void
blox_gtk_init ()
{
  int argc = 1;
  gchar *argvArray[] = { "gst", NULL };
  gchar **argv = argvArray;
  gtk_init (&argc, &argv);

  h_callback = _blox_create_callback_hash ();
}

void
my_gtk_main_iteration ()
{
  while (gtk_events_pending ())
    pending_quit_count += gtk_main_iteration_do (TRUE);
}

void
my_gtk_main_iteration_do (gboolean blocking)
{
  while (gtk_events_pending ())
    pending_quit_count += gtk_main_iteration_do (blocking);
}

gboolean
should_quit ()
{
  if (!pending_quit_count)
    return FALSE;

  pending_quit_count--;
  return TRUE;
}

void
widget_set_flags(GtkWidget *widget, int flags)
{
  GTK_WIDGET_SET_FLAGS (widget, flags);
}

void
widget_unset_flags(GtkWidget *widget, int flags)
{
  GTK_WIDGET_UNSET_FLAGS (widget, flags);
}

OOP
widget_get_property(GObject *anObject, const char *aProperty)
{
  GParamSpec *spec;
  GValue result;
  GObject *obj;

  obj = G_OBJECT (anObject);
  spec = g_object_class_find_property (G_OBJECT_GET_CLASS(obj), aProperty);
  g_value_init (&result, spec->value_type);
  g_object_get_property (obj, aProperty, &result);
  return (convert_g_value_to_oop (&result));
}

void
widget_set_property(GObject *anObject, const char *aProperty, OOP aValue)
{
  GParamSpec *spec;
  GObject *obj;
  GValue value;

  obj = G_OBJECT (anObject);
  spec = g_object_class_find_property (G_OBJECT_GET_CLASS(obj), aProperty);
  g_value_init (&value, spec->value_type);
  fill_g_value_from_oop (&value, aValue);
  g_object_set_property (obj, aProperty, &value);
}

OOP
container_get_child_property(GtkContainer *aParent, GtkWidget *aChild, const char *aProperty)
{
  GParamSpec *spec;
  GValue result;

  g_return_val_if_fail (GTK_WIDGET (aParent) ==
		        gtk_widget_get_parent (GTK_WIDGET (aChild)),
			_blox_vm_proxy->nilOOP);

  spec = gtk_container_class_find_child_property (G_OBJECT_GET_CLASS (aParent),
					          aProperty);

  g_value_init(&result, spec->value_type);
  gtk_container_child_get_property(aParent, aChild, aProperty, &result);
  return (convert_g_value_to_oop (&result));
}

void
container_set_child_property(GtkContainer *aParent, GtkWidget *aChild, const char *aProperty, OOP aValue)
{
  GParamSpec *spec;
  GValue value;

  g_return_if_fail (GTK_WIDGET (aParent) ==
		    gtk_widget_get_parent (GTK_WIDGET (aChild)));

  spec = gtk_container_class_find_child_property (G_OBJECT_GET_CLASS (aParent),
						  aProperty);

  g_value_init (&value, spec->value_type);
  fill_g_value_from_oop (&value, aValue);
  gtk_container_child_set_property(aParent, aChild, aProperty, &value);
}


void
gst_initModule (proxy)
     VMProxy *proxy;
{
  _blox_vm_proxy = proxy;
  _blox_vm_proxy->defineCFunc ("bloxGtkInit", blox_gtk_init);
  _blox_vm_proxy->defineCFunc ("bloxGtkConnectSignal", connect_signal);
  _blox_vm_proxy->defineCFunc ("bloxGtkMainIteration", my_gtk_main_iteration);
  _blox_vm_proxy->defineCFunc ("bloxGtkMainIterationDo", my_gtk_main_iteration_do);
  _blox_vm_proxy->defineCFunc ("bloxGtkShouldQuit", should_quit);
  _blox_vm_proxy->defineCFunc ("bloxGtkGetProperty", widget_get_property);
  _blox_vm_proxy->defineCFunc ("bloxGtkSetProperty", widget_set_property);
  _blox_vm_proxy->defineCFunc ("bloxGtkGetChildProperty", container_get_child_property);
  _blox_vm_proxy->defineCFunc ("bloxGtkSetChildProperty", container_set_child_property);
  _blox_vm_proxy->defineCFunc ("bloxGtkSetFlags", widget_set_flags);
  _blox_vm_proxy->defineCFunc ("bloxGtkUnsetFlags", widget_unset_flags);

  _blox_vm_proxy->defineCFunc ("gtk_placer_get_type", gtk_placer_get_type);
  _blox_vm_proxy->defineCFunc ("gtk_placer_new", gtk_placer_new);
  _blox_vm_proxy->defineCFunc ("gtk_placer_put", gtk_placer_put);
  _blox_vm_proxy->defineCFunc ("gtk_placer_move", gtk_placer_move);
  _blox_vm_proxy->defineCFunc ("gtk_placer_resize", gtk_placer_resize);
  _blox_vm_proxy->defineCFunc ("gtk_placer_move_rel", gtk_placer_move_rel);
  _blox_vm_proxy->defineCFunc ("gtk_placer_resize_rel", gtk_placer_resize_rel);
  _blox_vm_proxy->defineCFunc ("gtk_placer_set_has_window", gtk_placer_set_has_window);
  _blox_vm_proxy->defineCFunc ("gtk_placer_get_has_window", gtk_placer_get_has_window);

}
